const grammar = require('./grammar+');
const axiom = [
    (A) => ({
        ATOM: 'axiom 0',
        TYPE: grammar.stringify({
            op: '->',
            left: A,
            right: A
        })
    }),
    (A, B) => ({
        ATOM: 'axiom 1',
        TYPE: grammar.stringify({
            op: '->',
            left: A,
            right: {op: '->', left: B, right: A}
        })
    }),
    (A, B, C) => ({
        ATOM: 'axiom 2',
        TYPE: grammar.stringify({
            op: '->',
            left: {
                op: '->',
                left: A,
                right: {op: '->', left: B, right: C}
            },
            right: {
                op: '->',
                left: {op: '->', left: A, right: B},
                right: {op: '->', left: A, right: C}
            }
        })
    })
]

function FV(x, cl) {
    if(cl.TYPE===x) return true;
    if(cl.ATOM) return false;
    return FV(x, cl.left) || FV(x, cl.right);
}

function define_lamda(x, cl) {
    if(cl.TYPE===x) return axiom[0](x);
    let isAtom = cl.ATOM;
    if(!isAtom && !FV(x, cl.left)) {
        if(cl.right.TYPE===x) return cl.left;
        if(!FV(x, cl.right)) isAtom = true;
    }
    return isAtom ? {
        TYPE: grammar.stringify({op: '->', left: x, right: cl.TYPE}),
        left: axiom[1](cl.TYPE, x),
        right: cl
    } : {
        TYPE: grammar.stringify({op: '->', left: x, right: cl.TYPE}),
        left: {
            TYPE: grammar.stringify({
                op: '->',
                left: {op: '->', left: x, right: cl.right.TYPE},
                right: {op: '->', left: x, right: grammar.right(cl.left.TYPE)}
            }),
            left: axiom[2](x, cl.right.TYPE, grammar.right(cl.left.TYPE)),
            right: define_lamda(x, cl.left)
        },
        right: define_lamda(x, cl.right)
    }
}

function iks(lamda) {
    switch(lamda.op) {
    case 'lamda':
        return define_lamda(lamda.left, iks(lamda.right));
    case 'apply':
        let left = iks(lamda.left),
            right = iks(lamda.right);
        return {TYPE: grammar.right(left.TYPE), left, right}
    default:
        return {ATOM: 'FV', TYPE: lamda}
    }
}

module.exports = iks;
