const grammar = require('./grammar+');
const Queue   = require('./PriorityQueue');
const lamda   = require('./lamda');

function deduce(p) {
    let queue = new Queue(function(a, b) {
        for(let i=0; i<4; i++)
            if(a.priority[i]!==b.priority[i])
                return a.priority[i]<b.priority[i];
        return false;
    });
    queue.push(p);
    let target;
    while(true) {
        let expr = queue.pop();
        if(!expr) return null;
        if(expr.conclusion instanceof Array) {
            if(!expr.conclusion.length) {
                delete expr.used[target];
                return expr;
            }
            let r = deduce({
                priority: [...expr.priority],
                premise: {...expr.premise},
                used: {...expr.used},
                conclusion: expr.conclusion.shift(),
                lamda: null,
                apply: null
            });
            if(!r) continue;
            expr.priority = r.priority;
            for(let s in r.premise) {
                if(expr.premise[s]) continue;
                switch(r.premise[s].op) {
                case 'axiom 3':
                case 'axiom 4':
                case 'axiom 5':
                case 'axiom 6':
                case 'axiom 7':
                case 'axiom 8':
                case 'axiom 9':
                case 'axiom 10':
                    expr.premise[s] = r.premise[s];
                }
            }
            expr.apply = lamda.apply(expr.apply, lamda.lamda(r.lamda, r.apply));
            queue.push(expr);
        }
        else {
            while(expr.conclusion.op==='->') {
                let s = grammar.stringify(expr.conclusion.left);
                expr.priority[0]++;
                if(!expr.premise[s]) {
                    expr.premise[s] = expr.conclusion.left;
                    expr.used = {}
                }
                expr.conclusion = expr.conclusion.right;
                expr.lamda = lamda.lamda(expr.lamda, {op: 'lamda', left: s});
            }
            target = grammar.stringify(expr.conclusion);
            if(expr.used[target]) continue;
            switch(expr.conclusion.op) {
            case '*':
                grammar.axiom5(expr);
                break;
            case '+':
                grammar.axiom6(expr);
                grammar.axiom7(expr);
                break;
            default:
                if(expr.conclusion==='F') break;
                grammar.axiom9(expr);
                grammar.axiom10(expr);
            }
            for(let s in expr.premise)
                switch(expr.premise[s].op) {
                case '*':
                    grammar.axiom3(expr.premise, s);
                    grammar.axiom4(expr.premise, s);
                    break;
                case '+':
                    grammar.axiom8(expr.premise, s, expr.conclusion);
                }
            for(let s in expr.premise) {
                let q = grammar.find(expr.premise[s], expr.conclusion);
                if(!q) continue;
                let r = {
                    priority: [...expr.priority],
                    premise: {...expr.premise},
                    used: Object.assign({[target]: true}, expr.used),
                    conclusion: q,
                    lamda: lamda.clone(expr.lamda),
                    apply: s
                }
                r.priority[0]++;
                switch(r.premise[s].op) {
                case 'axiom 3':
                case 'axiom 4':
                case 'axiom 5':
                case 'axiom 6':
                case 'axiom 7':
                case 'axiom 8':
                    r.priority[1]++; break;
                case 'axiom 9':
                    r.priority[2]++; break;
                case 'axiom 10':
                    r.priority[3]++; break;
                }
                queue.push(r);
            }
        }
    }
}

module.exports = deduce;
