const grammar = require('./grammar');

function stringify(p) {
    switch(typeof p) {
    case 'string':
        return p;
    default:
        if(!p.formula) {
            let op = p.op;
            switch(op) {
            case 'axiom 3':
            case 'axiom 4':
            case 'axiom 5':
            case 'axiom 6':
            case 'axiom 7':
            case 'axiom 8':
            case 'axiom 9':
            case 'axiom 10':
                op = '->';
            }
            p.formula = '('+stringify(p.left)+op+stringify(p.right)+')';
        }
        return p.formula;
    }
}

function find(premise, conclusion) {
    let pstr = stringify(premise),
        cstr = stringify(conclusion);
    if(pstr.length < cstr.length) return null;
    if(pstr===cstr) return []
    switch(premise.op) {
    case 'axiom 3':
    case 'axiom 4':
    case 'axiom 5':
    case 'axiom 6':
    case 'axiom 7':
    case 'axiom 8':
    case 'axiom 9':
    case 'axiom 10':
    case '->':
        let p = find(premise.right, conclusion);
        if(p) p.unshift(premise.left);
        return p;
    default:
        return null;
    }
}

function right(s) {
    let p = grammar.parse(s, {startRule: 'Imply'});
    return stringify(p.right);
}

function axiom3(premise, s) {
    let axiom3 = {op: 'axiom 3', left: premise[s], right: premise[s].left}
    if(!premise[stringify(axiom3)])
        premise[stringify(axiom3)] = axiom3;
}

function axiom4(premise, s) {
    let axiom4 = {op: 'axiom 4', left: premise[s], right: premise[s].right}
    if(!premise[stringify(axiom4)])
        premise[stringify(axiom4)] = axiom4;
}

function axiom5(expr) {
    let axiom5 = {
        op: 'axiom 5',
        left: expr.conclusion.left,
        right: {op: '->', left: expr.conclusion.right, right: expr.conclusion}
    }
    if(!expr.premise[stringify(axiom5)])
        expr.premise[stringify(axiom5)] = axiom5;
}

function axiom6(expr) {
    let axiom6 = {op: 'axiom 6', left: expr.conclusion.left, right: expr.conclusion}
    if(!expr.premise[stringify(axiom6)])
        expr.premise[stringify(axiom6)] = axiom6;
}

function axiom7(expr) {
    let axiom7 = {op: 'axiom 7', left: expr.conclusion.right, right: expr.conclusion}
    if(!expr.premise[stringify(axiom7)])
        expr.premise[stringify(axiom7)] = axiom7;
}

function axiom8(premise, s, conclusion) {
    let axiom8 = {
        op: 'axiom 8',
        left: {op: '->', left: premise[s].left, right: conclusion},
        right: {
            op: '->',
            left: {op: '->', left: premise[s].right, right: conclusion},
            right: {op: '->', left: premise[s], right: conclusion}
        }
    }
    if(!premise[stringify(axiom8)])
        premise[stringify(axiom8)] = axiom8;
}

function axiom9(expr) {
    let axiom9 = {op: 'axiom 9', left: 'F', right: expr.conclusion}
    if(!expr.premise[stringify(axiom9)])
        expr.premise[stringify(axiom9)] = axiom9;
}

function axiom10(expr) {
    let axiom10 = {
        op: 'axiom 10',
        left: {
            op: '->',
            left: {op: '->', left: expr.conclusion, right: 'F'},
            right: expr.conclusion
        },
        right: expr.conclusion
    }
    if(!expr.premise[stringify(axiom10)])
        expr.premise[stringify(axiom10)] = axiom10;
}

module.exports = Object.assign({
    stringify,
    find,
    right,
    axiom3,
    axiom4,
    axiom5,
    axiom6,
    axiom7,
    axiom8,
    axiom9,
    axiom10
}, grammar);
